#!/usr/bin/env python
# coding: utf-8
# $Id$
import re
import math
import numpy as numpy
from functools import reduce
import kaisekiInterFace
from MyViewGlobal import logger

def helptext(title_name, script_name, data_source, synopsis, author_name, contact_to, plot_type, proc_summary, module_option):
    return kaisekiInterFace.helptext( title_name, script_name, data_source, synopsis, author_name, contact_to, plot_type, proc_summary, module_option )

class eg4dInterFace( kaisekiInterFace.kaisekiInterFace ):
    pattern   = '\Alhdmse[12]_shear\Z'            # lhdmse1_dfit, lhdmse2_dfit

    def __init__(self, data_id, title_name, data_source, tuple_ntavg, **kwargs):
        """
        kwargs:
            option: option string for trace
            cache: myview cache object
            cachebutton: state of button cache on myView
                         "CACHE":
                         "ONLINE":
                         "OFFLINE":
            temppath: temp path set with myView dialog
            cmdpath: command path set with myView dialog
        if there is some trouble to make instance, raise "Error"
        """
        super().__init__(data_id, title_name, data_source, **kwargs)
        self.R  = 0.0
        self.tm = 3.0
        self.parse_option(self.option)

    def parse_option(self, option):
        t = 3.0
        R = 0.0
        for p in [o.strip() for o in option.split(';')]:
            keyvalue = [item.strip() for item in p.split('=')]
            if len(keyvalue) == 2:
                key, value = keyvalue
                if key.lower() == 't':
                    try:
                        t = float(value)
                    except ValueError:
                        t = 0.0
                if key.lower() == 'r':
                    try:
                        R = float(value)
                    except ValueError:
                        R = 0.0
        self.R  = R
        self.tm = t

    def load(self, **kwargs):
      try:
        super().load( **kwargs )
        if self.flgloaded:
            return
            
        if self.cachepath:
            self.eg = EG_4D(self.cachepath)
            self.cache.releasefile(self.cachepath)
            self.flgloaded = True
            
        else:
            self.eg = None
      except Exception as e:
          print("load error =>", e)

    def varsx(self, **kwargs):
        """
        return list of candidates of variables for x-axis
        kwargs:
            option: option string for trace
            varx: varx set within trace dialog
            vary: vary set within trace dialog
            varz: varz set within trace dialog
        """
        lst = []
        self.load(**kwargs)
        if not self.eg:
            pass
        else :
            lst = ['dgamma', 'dgamma/ds']
            
        return lst

    def varsy(self, **kwargs):
        """
        return list of candidates of variables for y-axis or y2-axis
        kwargs:
            option: option string for trace
            varx: varx set within trace dialog
            vary: vary set within trace dialog
            varz: varz set within trace dialog
        """
        lst = []
        self.load(**kwargs)
        if not self.eg:
            pass
        else:
            ptype = kwargs.get('type', None)
            if ptype == 'xyz':
                pass
                
            else :
                varx  = kwargs.get('varx')
                if varx == 'dgamma':
                    lst = ['diota', 'diota_fit', 'diota_line', 'diota_line_oneip', 
                           'diota_ipindex', 'diota_fit_ipindex']
                else:
                    lst = ['diota/ds', 'diota/ds_fit', 'diota/ds_line', 
                           'diota/ds_ipindex', 'diota/ds_fit_ipindex']
                    
        return lst

    def varsz(self, **kwargs):
        """
        return list of candidates of variables for z-axis
        kwargs:
            option: option string for trace
            varx: varx set within trace dialog
            vary: vary set within trace dialog
            varz: varz set within trace dialog
        """
        return []

    def indices_of_current_prms(self):
        ips = self.eg.dims(1)
        pfs = self.eg.dims(2)
        lst = []
        for i, ip in enumerate(ips):
            lst.append((ip, i, 2))
            
        """
        for i, pf in enumerate(pfs):
            lst.append((pf, i, 1))
        """
        return lst # (ip or pf, dim_idx for trace)

    def get_time_R(self, tm, R):
        tlist = self.eg.dims(0)
        rlist = self.eg.dims(3)
        tidx, t = kaisekiInterFace.FindNear(tm, tlist)
        ridx, r = kaisekiInterFace.FindNear(R, rlist)
        return t, tidx, r, ridx

    def indexs(self, **kwargs):
        """
        return list of indices for varx, vary, varz and option.
        kwargs:
            option: option string for trace
            varx: varx set within trace dialog
            vary: vary set within trace dialog
            varz: varz set within trace dialog
        """
        lst = []
        self.load(**kwargs)
        if not self.eg:
            pass
        else:
            ptype = kwargs.get('type', None)
            if ptype == 'xyz':
                pass
                
            else:
                vary = kwargs.get('vary')
                if vary in list(self.ip_pf_independent_values.keys()):
                    pass
                else:
                    lst = ["%d_%g" % (i, tpl[0])
                           for i, tpl in enumerate(self.indices_of_current_prms())]
                    
        return lst

    def dataset(self, **kwargs):
        """
        return list of dataset for varx, vary, varz, index and option.
        kwargs:
            option: option string for trace
            varx: varx set within trace dialog
            vary: vary set within trace dialog
            varz: varz set within trace dialog
        dataset := [title, x, xe, y, ye, z, ze]
        x = [x1, x2, x3, ... , xn]
        y = [y1, y2, y3, ... , ym]
        z = [[z(x1, y1), z(x2, y1), z(x3, y1), ... , z(xn, y1)],
             [z(x1, y2), z(x2, y2), z(x3, y2), ... , z(xn, y2)],
             [z(x1, y3), z(x2, y3), z(x3, y3), ... , z(xn, y3)],
             ......
             [z(x1, ym), z(x2, ym), z(x3, ym), ... , z(xn, ym)]]
        len(z) = len(y)
        len(z[i]) = len(x)
        """
        self.load(**kwargs)
        
        title = '%s@%s' % (self.diagname, self.dataid)
        if not self.eg:
            # if there is no data
            return [title+'_No Data', [], [], [], [], [], []]
            
        else:
            self.refresh_index(kwargs)   # add 2023.7.21 yoshida
            ptype = kwargs.get('type')
            if ptype == 'xyz':
                return [title+'_N/A', [], [], [], [], [], []]
                
            else:
                vary = kwargs.get('vary')
                if vary in list(self.ip_pf_independent_values.keys()):
                    return self.plot_points_all(**kwargs)
                    
                if vary in list(self.vdic.keys()):
                    return self.plot_points(**kwargs)
                    
                return self.fit_line(**kwargs)

    def plot_points_all(self, **kwargs):
        varx, vary = kaisekiInterFace.KeywordGet(kwargs, ('varx', 'vary'))
        xdata, xedata, ydata, yedata, zdata, zedata = [], [], [], [], [], []

        xname, flg = self.vdic.get(varx)
        yname = self.ip_pf_independent_values.get(vary)
        if vary[-3:] == 'fit':
            select_fitting_value = True
        else:
            select_fitting_value = False
            
        xvidx = self.eg.vidx(xname)
        xunit = self.eg.valunits[xvidx]
        yvidx = self.eg.vidx(yname)
        yunit = self.eg.valunits[yvidx]
        flg_vidx = self.eg.vidx('flag')

        tm = self.tm
        R  = self.R
        tm, tidx, R, ridx = self.get_time_R(tm, R)
        ips = self.eg.dims(1)
        pfs = self.eg.dims(2)

        for ipidx, ip in enumerate(ips): # ip scan
            didx = 2 # pf profile
            idxs = [tidx, ipidx, ridx]
            flgdata = self.eg.trace_of_vidx(didx, idxs, flg_vidx)
            xtrace  = self.eg.trace_of_vidx(didx, idxs, xvidx)
            ytrace  = self.eg.trace_of_vidx(didx, idxs, yvidx)
            xdata   = xdata + [v for v, f in zip(xtrace, flgdata) 
                      if int(f)!=0 or (not select_fitting_value)]
            ydata   = ydata + [v for v, f in zip(ytrace, flgdata) 
                      if int(f)!=0 or (not select_fitting_value)]
            
        for pfidx, pf in enumerate(pfs): # ip scan
            didx = 1 # pf profile
            idxs = [tidx, pfidx, ridx]
            flgdata = self.eg.trace_of_vidx(didx, idxs, flg_vidx)
            xtrace  = self.eg.trace_of_vidx(didx, idxs, xvidx)
            ytrace  = self.eg.trace_of_vidx(didx, idxs, yvidx)
            xdata   = xdata + [v for v, f in zip(xtrace, flgdata) 
                      if int(f)!=0 or (not select_fitting_value)]
            ydata   = ydata + [v for v, f in zip(ytrace, flgdata)
                      if int(f)!=0 or (not select_fitting_value)]
            
        if varx == 'dgamma':
            txt = "%s@%s_tan(dgamma)" % (self.diagname, self.dataid)
            xdata = [math.tan(g*math.pi/180.0) for g in xdata]
        else:
            txt = "%s@%s_dgamma/dx" % (self.diagname, self.dataid)
            
        title = '%s_vs_%s_%gsec_%gm' % (txt, vary, tm, R)
        return [title, xdata, xedata, ydata, yedata, zdata, zedata]

    def plot_points(self, **kwargs):
        varx, vary, index = kaisekiInterFace.KeywordGet(kwargs, ('varx', 'vary', 'index'))
        xdata, xedata, ydata, yedata, zdata, zedata = [], [], [], [], [], []

        idx = int(index.split('_')[0])
        tpl = self.indices_of_current_prms()

        if idx >= len(tpl):
            idx = len(tpl) - 1
            
        ip_or_pf, ipidx, didx = tpl[idx]

        xname, flg = self.vdic.get(varx)
        yname, fit_used_only = self.vdic.get(vary)
        xvidx = self.eg.vidx(xname)
        xunit = self.eg.valunits[xvidx]
        yvidx = self.eg.vidx(yname)
        yunit = self.eg.valunits[yvidx]
        flg_vidx = self.eg.vidx('flag')

        tm = self.tm
        R = self.R
        tm, tidx, R, ridx = self.get_time_R(tm, R)

        idxs = [tidx, ipidx, ridx]
        xdata = self.eg.trace_of_vidx(didx, idxs, xvidx)
        if varx == 'dgamma':
            txt   = "%s@%s_tan(dgamma)" % (self.diagname, self.dataid)
            xdata = [math.tan(g*math.pi/180.0) for g in xdata]
        else:
            txt = "%s@%s_dgamma/dx" % (self.diagname, self.dataid)
            
        ydata   = self.eg.trace_of_vidx(didx, idxs, yvidx)
        flgdata = self.eg.trace_of_vidx(didx, idxs, flg_vidx)

        if fit_used_only:
            xdata = [x for x, f in zip(xdata, flgdata) if int(f)!=0]
            ydata = [y for y, f in zip(ydata, flgdata) if int(f)!=0]
            
        if didx == 1:
            ip_or_pf_str = "pf=%d" % ip_or_pf
        else:
            ip_or_pf_str = "ip=%d" % ip_or_pf
            
        title = '%s_vs_%s_%gsec_%gm_%s' % (txt, vary, tm, R, ip_or_pf_str)

        return [title, xdata, xedata, ydata, yedata, zdata, zedata]

    def fit_line(self, **kwargs):
        varx, vary, index = kaisekiInterFace.KeywordGet(kwargs, ('varx', 'vary', 'index'))
        xdata, xedata, ydata, yedata, zdata, zedata = [], [], [], [], [], []

        idx = int(index.split('_')[0])
        ip_or_pf, ipidx, didx = self.indices_of_current_prms()[idx]

        aname, bname = self.fitline_dic.get(vary)

        xname, flg = self.vdic.get(varx)
        xvidx = self.eg.vidx(xname)
        reffvidx = self.eg.vidx('reff')

        avidx = self.eg.vidx(aname)
        bvidx = self.eg.vidx(bname)

        tm = self.tm
        R = self.R
        tm, tidx, R, ridx = self.get_time_R(tm, R)

        idxs = [tidx, ipidx, ridx]
        xdata = self.eg.trace_of_vidx(didx, idxs, xvidx)
        reffs = self.eg.trace_of_vidx(didx, idxs, reffvidx)
        adata = self.eg.trace_of_vidx(didx, idxs, avidx)
        bdata = self.eg.trace_of_vidx(didx, idxs, bvidx)
        Rax = float(self.eg.keyvalue.get('Bax'))
        a = adata[0]
        b = bdata[0]

        if varx == 'dgamma':
            txt = "%s@%s_tan(dgamma)" % (self.diagname, self.dataid)
            xdata = [math.tan(g*math.pi/180.0) for g in xdata]
        else:
            txt = "%s@%s_dgamma/dx" % (self.diagname, self.dataid)
            
        ydata = [(a * x + b) for x in xdata]

        if didx == 1:
            ip_or_pf_str = "pf=%d" % ip_or_pf
        else:
            ip_or_pf_str = "ip=%d" % ip_or_pf
            
        title = '%s_vs_%s_%gsec_%gm_%s' % (txt, vary, tm, R, ip_or_pf_str)
        return [title, xdata, xedata, ydata, yedata, zdata, zedata]

class EG_4D(object):
    def __init__(self, filepath):
        """
        self.data[idx]
        iline = self.line_index()
        size = self.dimno + self.valno
        idx = self.line_size * iline + vidx
        """
        self.keyvalue = {}
        self.filepath = filepath
        self.load(filepath)

    def load(self, filepath):
        zf, fp = kaisekiInterFace.OpenFile(filepath)
        l = fp.readline()
        while l:
            l = l.strip()
            if len(l) == 0:
                l = fp.readline()
                continue
            if l[0] == '#':
                clm = l[1:].split('=')
                if len(clm) == 2:
                    self.keyvalue[clm[0].strip()] = clm[1].strip()
                    # print '%s' % clm[1].strip()
                l = fp.readline()
                continue
            break
        cdimnames = self.keyvalue.get('DimName')
        cdimunits = self.keyvalue.get('DimUnit')
        cdimno    = self.keyvalue.get('DimNo')
        cdimsize  = self.keyvalue.get('DimSize')
        cvalno    = self.keyvalue.get('ValNo')
        cvalnames = self.keyvalue.get('ValName')
        cvalunits = self.keyvalue.get('ValUnit')
        dimno     = int(cdimno)
        valno     = int(cvalno)
        line_size = dimno + valno
        dimsizes  = [int(c.strip()) for c in cdimsize.split(',')]
        data_size = line_size * reduce((lambda x, y: x * y), dimsizes, 1)

        # allocate memory
        self.data = numpy.zeros(data_size)

        # make Headder info
        self.dimnames = [c.strip("' ") for c in cdimnames.split(',')]
        self.dimunits = [c.strip("' ") for c in cdimunits.split(',')]
        self.valnames = [c.strip("' ") for c in cvalnames.split(',')]
        self.valunits = [c.strip("' ") for c in cvalunits.split(',')]
        self.dimno = dimno
        self.valno = valno
        self.line_size = line_size
        self.dimsizes  = dimsizes
        self.weights = kaisekiInterFace.CalcWeights(self.dimsizes[1:])

        idx = 0
        while l:
            l = l.strip()
            if len(l) == 0:
                l = fp.readline()
                continue
                
            for c in re.split(r',\s*|\s+',l):
                if c != "":
                    try:
                        self.data[idx] = float(c)
                    except ValueError:
                        self.data[idx] = numpy.nan
                    idx += 1
                    
            l = fp.readline()
            
        kaisekiInterFace.CloseFile(zf, fp)

    def line_index(self, *dimidxs):
        weighted_indexs = [i * w for i, w in zip(dimidxs, self.weights)]
        return reduce((lambda x, y: x + y), weighted_indexs, 0)

    def get_line(self, *dimidxs):
        idx = self.line_size * self.line_index(*dimidxs)
        return self.data[idx:idx + self.line_size]

    def dims(self, dimidx):
        pre  = [0] * dimidx
        post = [0] * (self.dimno-dimidx)
        pos  = [pre+[i]+post for i in range(self.dimsizes[dimidx])]
        return [self.get_line(*p)[dimidx] for p in pos]

    def vdata(self, idxv, *dimidxs):
        return self.get_line(*dimidxs)[self.dimno + idxv]

    def vidx(self, name):
        if name in self.valnames:
            return self.valnames.index(name)
        else:
            return -1

    def trace_of_vidx(self, dimidx, other_indices, vidx):
        """
        dimidx: idx of variable dims for trace
        ex)
          0: time trace
          1: radial profile
        """
        dimno = self.dimno
        dimsize = self.dimsizes[dimidx]
        pos = [other_indices[:dimidx] + [i] + other_indices[dimidx:] for i in range(dimsize)]
        return [self.get_line(*p)[self.dimno + vidx] for p in pos]

    def trace_of(self, dimidx, idx, name):
        """
        dimidx: idx of variable dims for trace
        ex)
          0: time trace
          1: radial profile
        """
        vidx = self.vidx(name)
        if vidx == -1:
            return []
        dimno   = self.dimno
        dimsize = self.dimsizes[dimidx]
        pos = [other_indices[:dimidx] + [i] + other_indices[dimidx:] for i in range(dimsize)]
        return [self.get_line(*p)[self.dimno + vidx] for p in pos]
