#!/usr/bin/env python
# coding: utf-8
# $Id$

import re
import math
import os
from MyViewGlobal import logger
from egdbcommon import *
import numpy

class egdb2d():
    """
    egdb() is pure python module for parsing EGDB formatted file.
    egdb()はEGDB形式のファイルを分解して読み込むためのモジュールです。

    if you want to open 'temp.dat', make instance of the file with edgb(), 
    and then call readFile() method.
    'temp.dat'というファイルを開きたい場合は、egdb()を使ってインスタンスを作成し、
    readFile()メソッドを実行します。
    ex)
        egfile = egdb('temp.dat')
        egfile.readFile()

    """
    def __init__(self, fname=None, delim=',', tdic={}):
        """
        fname    full path of the file you want to open.
                 開きたいファイルのフルパスを設定します。
        delim    delimiter of the data.
                 データの区切り文字を指定します。
        tdic     the dectionary for translation from non-numeric value 
                 to the value in the dictionary.
                 非数値の文字を、辞書を使って変換します。
                 ex) tdic = {'+Inf': '99999.9', '-Inf': '-99999.9'}
        """
        self.fileName = fname
        self.delim = delim
        self.parameters = ''
        self.comments = ''
        self.Name = ''
        self.ShotNo = ''
        self.Date = ''
        self.DimNo=0
        self.DimSize=[0]
        self.DimName=""
        self.DimUnit=""
        self.ValNo=0
        self.ValNames=[]
        self.ValUnits=[]
        self.parametersDict = {}
        self.parametersFlg = False
        self.commentsFlg = False
        self.dataFlg = False
        
        self.dimdata = []
        self.data = []
        
        self.reobjTitle = re.compile(r'\s*\[(.+)\]\s*')
        self.reobjValue = re.compile(r'(.+?)\s*=\s*(.+)')

    def parseLine(self, line):
        """
        parsing a line of headder and getting the size of the data.
        ヘッダを解析して、メモリサイズを得て、適切なサイズのリストを作成する
        """
        #reobjTitle = re.compile(r'\s*\[(.+)\]\s*')
        m = self.reobjTitle.match(line)
        if m:
            idx = m.groups()[0].upper()
            if 'PARAMETERS' == idx:
                self.parametersFlg = True
                self.commentsFlg = False
                self.dataFlg = False 
            if 'COMMENTS' == idx:
                self.parametersFlg = False
                self.commentsFlg = True
                self.dataFlg = False 
            if 'DATA' == idx:
                self.parametersFlg = False
                self.commentsFlg = False 
                self.dataFlg = True
            return

        #reobjValue = re.compile(r'(.+)\s*=\s*(.+)')
        m = self.reobjValue.match(line)
        if m:
            #print m.groups()
            nm = m.groups()[0].upper()
            nm = nm.strip()
            vl = m.groups()[1]
            if 'NAME' == nm:
                clm = vl.strip()
                clm = clm.strip('\'')
                self.Name = clm
            if 'SHOTNO' == nm:
                clm = vl.strip()
                self.ShotNo = clm
            if 'DATE' == nm:
                clm = vl.strip()
                clm = clm.strip('\'')
                self.Date = clm

            if 'DIMNO' == nm:
                self.DimNo = int(vl)

            if 'DIMSIZE' == nm:
                for i in range(self.DimNo):
                    clm = vl.split(',')
                    self.DimSize[0] = int(clm[0])

            if 'DIMNAME' == nm:
                clm = vl.split(',')
                clmd = clm[0].strip()
                clmd = clmd.strip('\'')
                clmd = clmd.strip()
                self.DimName = clmd

            if 'DIMUNIT' == nm:
                clm = vl.split(',')
                clmd = clm[0].strip()
                clmd = clmd.strip('\'')
                clmd = clmd.strip()
                self.DimUnit = clmd
            if 'VALNO' == nm:
                self.ValNo = int(vl)
                for i in range(self.ValNo):
                    self.data.append([])
            if 'VALNAME' == nm:
                for i in range(self.ValNo):
                    clm = vl.split(',')
                    clmd = clm[i].strip()
                    clmd = clmd.strip('\'')
                    clmd = clmd.strip()
                    self.ValNames.append(clmd)
            if 'VALUNIT' == nm:
                for i in range(self.ValNo):
                    clm = vl.split(',')
                    clmd = clm[i].strip()
                    clmd = clmd.strip('\'')
                    clmd = clmd.strip()
                    self.ValUnits.append(clmd)
            self.parametersDict[nm] = vl

    def readFile(self):
        """
        reading file.
        ファイルの読み込みを実行する
        """
        if self.fileName[-3:].upper() == 'ZIP':
            # open zip file
            zf = zipfile.ZipFile(self.fileName, 'r')
            targetname = zf.namelist()[0]
            buff = zf.read(targetname)
            zf.close()
            lines = buff.decode('utf-8').split('\n')
        else:
            fp = open(self.fileName, 'r')
            lines = fp.readlines()
            fp.close()

        for i, l in enumerate(lines):
            l = l.strip('\n')
            l = l.strip('\r')
            l = l.strip()
            l = l.strip(',')
            if len(l) == 0: continue
            if l[0] == '#':
                self.parseLine(l[1:])
                if self.parametersFlg:
                    self.parameters = self.parameters+l+'\n'
                if self.commentsFlg:
                    self.comments = self.comments+l+'\n'
            else:
                if self.DimNo != 1:
                    logger.error("Not Single Dimension Data")
                    return False
                clm = l.split(self.delim)
                self.dimdata.append(float(clm[0]))
                for j, v in enumerate(clm[1:]):
                    try:
                        vm = float(v)
                    except ValueError:
                        #print v
                        vm = numpy.nan
                    self.data[j].append(float(vm))
        return True

    def getName(self):
        """
        return the name of diagnostic
        """
        return self.Name

    def getShotNo(self):
        """
        return the shot number written in the file.
        """
        return self.ShotNo

    def getDate(self):
        """
        return the DATE written in the file.
        """
        return self.Date

    def getDimNo(self):
        """
        return the number of dimensions.
        """
        return self.DimNo

    def getDimSize(self, dim):
        """
        return the size of dimensions.
        """
        return self.DimSize[dim]

    def getValNo(self):
        """
        return the number of variables (Values).
        """
        return self.ValNo
    
    def getParameters(self):
        """
        return the block of the parameters.
        パラメータブロックを得る
        """
        return self.parameters

    def getComments(self):
        """
        return the block of the comments.
        コメントを得る
        """
        return self.comments

    def getParameterByName(self, name):
        """
        return the parameter in the comment block indicated by 'name ='.
        """
        nameu = name.upper()
        return self.parametersDict.get(nameu)

    def value2idx(self, v):
        """
        get an index value of the dimension at the value near the argument value.
        あるdimの値vに近い場所のインデックス値を得る

        ex) idx = value2idx(3.4)

        """
        dlist = []
        for i, dimvalue in enumerate(self.dimdata):
            d = abs(v - dimvalue)
            dlist.append([d, i])
        dlist.sort()
        idx = dlist[0][1]
        return idx
    
    def idx2value(self, idx):
        """
        get the value at the idx of the dimmension.
        ある次元dimの実際の値をインデックス値から得る

        ex) time = idx2value(self, 0, 10)
        """
        v = self.dimdata[idx]
        return v

    def dims(self, dim):
        """
        get the list of value in the dimension.
        次元dimの値のリストをインデックス順に得る
        """
        return self.dimdata

    def dimname(self):
        """
        get the list of the dimenstion name. 
        次元の名前を得る
        """
        return self.DimName

    def dimunit(self):
        """
        get the list of the dimenstion names. 
        次元の単位を得る
        """
        return self.DimUnit

    def valnames(self):
        """
        get the list of the variable (Value) names. 
        変数の名前のリストを得る
        """
        return self.ValNames

    def valunits(self):
        """
        get the list of the units of variable (Value). 
        変数の単位のリストを得る
        """
        return self.ValUnits

    def valname2idx(self, name):
        """
        get the index of the variable (Value) by the name.
        変数(値)名から、インデックスを得る
        """
        vnlist = [vn.upper() for vn in self.ValNames]
        if name.upper() in vnlist:
            return vnlist.index(name.upper())
        else:
            return -1

    def idx2valname(self, idx):
        """
        get the variable (Value) name by the index.
        インデックスから変数(値)名を得る
        """
        return self.ValNames[idx]

    def idx2valunit(self, idx):
        """
        get the unit of the variable (Value) by the index.
        インデックスから変数(値)の単位を得る
        """
        return self.ValUnits[idx]


if __name__ == '__main__':
    os.unlink('temp.dat')
    egfile = igetfile('nb1pwr_temporal', 101911, 1, 'temp.dat')
    egfile = egdb2d('temp.dat')
    egfile.readFile()
    print((egfile.Name))
    print((egfile.ShotNo))
    print((egfile.Date))
    print((egfile.DimNo))
    print((egfile.DimSize))
    print((egfile.DimName))
    print((egfile.DimUnit))
    print((egfile.ValNo))
    print((egfile.ValNames))
    print((egfile.ValUnits))
    print((egfile.parametersDict))
    print((egfile.valname2idx('Ebeam_nb1')))
    print((egfile.valname2idx('Pport-through_nb1')))
    print((egfile.valname2idx('Ion-stsA_nb1')))
    print((egfile.valname2idx('Ion-stsB_nb1')))

    #vidx = egfile.valname2idx('')
    #power = egfile.data[1][pos]
    #    print idxt, t, pos, ti

    #print(egfile.dimdata)
    #print(egfile.data[0])
    #for t, v in zip(egfile.dimdata, egfile.data[0]):
    #    print("%f\t%f" % (t, v))
    #t = registeredTime('tsmap', 120000, 1)
    #print('t=',t)
    
