#!/usr/bin/env python
# coding: utf-8
# $Id$

import re
import math
import os
from MyViewGlobal import logger
from egdbcommon import *

    
class egdb3d():
    """
    egdb() is pure python module for parsing EGDB formatted file.
    egdb()はEGDB形式のファイルを分解して読み込むためのモジュールです。

    if you want to open 'temp.dat', make instance of the file with edgb(), 
    and then call readFile() method.
    'temp.dat'というファイルを開きたい場合は、egdb()を使ってインスタンスを作成し、
    readFile()メソッドを実行します。
    ex)
        egfile = egdb('temp.dat')
        egfile.readFile()

    """
    def __init__(self, fname=None, delim=',', tdic={}):
        """
        fname    full path of the file you want to open.
                 開きたいファイルのフルパスを設定します。
        delim    delimiter of the data.
                 データの区切り文字を指定します。
        tdic     the dectionary for translation from non-numeric value 
                 to the value in the dictionary.
                 非数値の文字を、辞書を使って変換します。
                 ex) tdic = {'+Inf': '99999.9', '-Inf': '-99999.9'}
        """
        self.fileName = fname
        self.delim = delim
        self.parameters = ''
        self.comments = ''
        self.Name = ''
        self.ShotNo = ''
        self.Date = ''
        self.DimNo=2
        self.DimSizes=[0, 0]
        self.DimNames=["", ""]
        self.DimUnits=["", ""]
        self.ValNo=0
        self.ValNames=[]
        self.ValUnits=[]
        self.parametersDict = {}
        self.parametersFlg = False
        self.commentsFlg = False
        self.dataFlg = False
        self.tdic = tdic
        self.dimdata = [[], []]
        self.data = []
        
        self.reobjTitle = re.compile(r'\s*\[(.+)\]\s*')
        self.reobjValue = re.compile(r'(.+?)\s*=\s*(.+)')

    def parseLine(self, line):
        """
        parsing a line of headder and getting the size of the data.
        ヘッダを解析して、メモリサイズを得て、適切なサイズのリストを作成する
        """
        #reobjTitle = re.compile(r'\s*\[(.+)\]\s*')
        m = self.reobjTitle.match(line)
        if m:
            idx = m.groups()[0].upper()
            if 'PARAMETERS' == idx:
                self.parametersFlg = True
                self.commentsFlg = False
                self.dataFlg = False 
            if 'COMMENTS' == idx:
                self.parametersFlg = False
                self.commentsFlg = True
                self.dataFlg = False 
            if 'DATA' == idx:
                self.parametersFlg = False
                self.commentsFlg = False 
                self.dataFlg = True
            return

        #reobjValue = re.compile(r'(.+)\s*=\s*(.+)')
        m = self.reobjValue.match(line)
        if m:
            #print m.groups()
            nm = m.groups()[0].upper()
            nm = nm.strip()
            vl = m.groups()[1]
            if 'NAME' == nm:
                clm = vl.strip()
                clm = clm.strip('\'')
                self.Name = clm
            if 'SHOTNO' == nm:
                clm = vl.strip()
                self.ShotNo = clm
            if 'DATE' == nm:
                clm = vl.strip()
                clm = clm.strip('\'')
                self.Date = clm

            if 'DIMNO' == nm:
                self.DimNo = int(vl)

            if 'DIMSIZE' == nm:
                for i in range(self.DimNo):
                    clm = vl.split(',')
                    self.DimSizes[0] = int(clm[0])
                    self.DimSizes[1] = int(clm[1])
                self.dimdata[0] = [0.0 for i in range(self.DimSizes[0])]
                self.dimdata[1] = [0.0 for i in range(self.DimSizes[1])]

            if 'DIMNAME' == nm:
                clm = vl.split(',')
                for i in range(self.DimNo):
                    clmd = clm[i].strip()
                    clmd = clmd.strip('\'')
                    clmd = clmd.strip()
                    self.DimNames[i] = clmd

            if 'DIMUNIT' == nm:
                clm = vl.split(',')
                for i in range(self.DimNo):
                    clmd = clm[i].strip()
                    clmd = clmd.strip('\'')
                    clmd = clmd.strip()
                    self.DimUnits[i] = clmd

            if 'VALNO' == nm:
                self.ValNo = int(vl)
                for i in range(self.ValNo):
                    self.data.append([])
                self.data = [0.0 for i in 
                        range(self.DimSizes[0] * self.DimSizes[1] * self.ValNo)]
            if 'VALNAME' == nm:
                for i in range(self.ValNo):
                    clm = vl.split(',')
                    clmd = clm[i].strip()
                    clmd = clmd.strip('\'')
                    clmd = clmd.strip()
                    self.ValNames.append(clmd)
            if 'VALUNIT' == nm:
                for i in range(self.ValNo):
                    clm = vl.split(',')
                    clmd = clm[i].strip()
                    clmd = clmd.strip('\'')
                    clmd = clmd.strip()
                    self.ValUnits.append(clmd)
            self.parametersDict[nm] = vl

    def readFile(self):
        """
        reading file.
        ファイルの読み込みを実行する
        """
        if self.fileName[-3:].upper() == 'ZIP':
            # open zip file
            zf = zipfile.ZipFile(self.fileName, 'r')
            targetname = zf.namelist()[0]
            buff = zf.read(targetname)
            zf.close()
            lines = buff.decode('utf-8').split('\n')
        else:
            fp = open(self.fileName, 'r')
            lines = fp.readlines()
            fp.close()

        
        c0 = 0
        c1 = 0
        flg = True # Check First for Dim(0)
        for i, l in enumerate(lines):
            l = l.strip('\n')
            l = l.strip(',')
            if len(l) == 0: continue
            if l[0] == '#':
                self.parseLine(l[1:])
                if self.parametersFlg:
                    self.parameters = self.parameters+l+'\n'
                if self.commentsFlg:
                    self.comments = self.comments+l+'\n'
            else:
                if self.DimNo != 2:
                    logger.error("Not Two Dimension Data")
                    return False
                clm = l.split(self.delim)

                if flg:
                    v = float(clm[0])
                    self.dimdata[0][c0] = v
                    c1 = 0
                    c0 = c0 + 1
                    flg = False

                if c1 >= self.DimSizes[1]:
                    v = float(clm[0])
                    self.dimdata[0][c0] = v
                    c1 = 0
                    c0 = c0 + 1

                self.dimdata[1][c1] = float(clm[1])
                c1 = c1 + 1
                
                for j, v in enumerate(clm[2:]):
                    idx = j + (c1-1)*self.ValNo + (c0-1)*self.DimSizes[1]*self.ValNo 
                    #print c0, c1, self.DimSizes[0], self.DimSizes[1], self.ValNo, idx
                    try:
                        self.data[idx] = float(v)
                    except ValueError:
                        v = self.tdic.get(v, None)
                        if v:
                            self.data[idx] = float(v)
                        else:
                            self.data[idx] = 0.0


        return True

    def position(self, idx_dim0, idx_dim1, idx_val):
        """
        return index of (idx_dim0, idx_dim1, idx_val)
        """
        idx = idx_val + idx_dim1*self.ValNo + idx_dim0*self.DimSizes[1]*self.ValNo 
        return idx

    def getName(self):
        """
        return the name of diagnostic
        """
        return self.Name

    def getShotNo(self):
        """
        return the shot number written in the file.
        """
        return self.ShotNo

    def getDate(self):
        """
        return the DATE written in the file.
        """
        return self.Date

    def getDimNo(self):
        """
        return the number of dimensions.
        """
        return self.DimNo

    def getDimSize(self, dim):
        """
        return the size of dimensions.
        """
        return self.DimSizes[dim]

    def getValNo(self):
        """
        return the number of variables (Values).
        """
        return self.ValNo
    
    def getParameters(self):
        """
        return the block of the parameters.
        パラメータブロックを得る
        """
        return self.parameters

    def getComments(self):
        """
        return the block of the comments.
        コメントを得る
        """
        return self.comments

    def getParameterByName(self, name):
        """
        return the parameter in the comment block indicated by 'name ='.
        """
        nameu = name.upper()
        return self.parametersDict.get(nameu)

    def value2idx(self, v, dim):
        """
        get an index value of the dimension at the value near the argument value.
        あるdimの値vに近い場所のインデックス値を得る

        ex) idx = value2idx(3.4)

        """
        dlist = []
        for i, dimvalue in enumerate(self.dimdata[dim]):
            d = abs(v - dimvalue)
            dlist.append([d, i])
        dlist.sort()
        idx = dlist[0][1]
        return idx
    
    def idx2value(self, idx, dim):
        """
        get the value at the idx of the dimmension.
        ある次元dimの実際の値をインデックス値から得る

        ex) time = idx2value(self, 0, 10)
        """
        v = self.dimdata[dim][idx]
        return v

    def dims(self, dim):
        """
        get the list of value in the dimension.
        次元dimの値のリストをインデックス順に得る
        """
        return self.dimdata[dim]

    def dimname(self, dim):
        """
        get the list of the dimenstion name. 
        次元の名前を得る
        """
        return self.DimNames[dim]

    def dimunit(self, dim):
        """
        get the list of the dimenstion names. 
        次元の単位を得る
        """
        return self.DimUnits[dim]

    def valnames(self):
        """
        get the list of the variable (Value) names. 
        変数の名前のリストを得る
        """
        return self.ValNames

    def valunits(self):
        """
        get the list of the units of variable (Value). 
        変数の単位のリストを得る
        """
        return self.ValUnits

    def valunit(self, vidx):
        """
        get the list of the unit of variable with validx(Value). 
        vidx 番目の変数の単位を得る
        """
        return self.ValUnits[vidx]

    def valname2idx(self, name):
        """
        get the index of the variable (Value) by the name.
        変数(値)名から、インデックスを得る
        """
        vnlist = [vn.upper() for vn in self.ValNames]
        if name.upper() in vnlist:
            return vnlist.index(name.upper())
        else:
            return -1

    def idx2valname(self, idx):
        """
        get the variable (Value) name by the index.
        インデックスから変数(値)名を得る
        """
        return self.ValNames[idx]

    def idx2valunit(self, idx):
        """
        get the unit of the variable (Value) by the index.
        インデックスから変数(値)の単位を得る
        """
        return self.ValUnits[idx]


if __name__ == '__main__':
    #os.unlink('temp.dat')
    #egfile = igetfile('tsmap', 115000, 1, 'temp.dat')
    #egfile = egdb2d('temp.dat')
    """
    eg = egdb3d('tsmap@123130.dat')
    eg.readFile()
    print '#',eg.Name
    print '#',eg.ShotNo
    print '#',eg.Date
    print '#',eg.DimNo
    print '#',eg.DimSizes
    print '#',eg.DimNames
    print '#',eg.DimUnits
    print '#',eg.ValNo
    print '#',eg.ValNames
    print '#',eg.ValUnits
    print '#',eg.parametersDict
    print '#',eg.valname2idx('Te')
    print '#',eg.valname2idx('dTe')
    print '#',eg.valname2idx('n_e')
    print '#',eg.valname2idx('dn_e')

    tlist = eg.dims(0)
    rlist = eg.dims(1)
    tidx = eg.value2idx(4.3, 0)
    ridx = eg.value2idx(3.6, 1)

    vidx = eg.valname2idx('Te')

    for i, r in enumerate(rlist):
        pos = eg.position(tidx, i, vidx)
        te = eg.data[pos]
        print "%g\t%g\t%g" % (tlist[tidx], r, te)
    """
    """
    for i, t in enumerate(tlist):
        pos = eg.position(i, ridx, vidx)
        te = eg.data[pos]
        print "%g\t%g\t%g" % (rlist[ridx], t, te)
    """
    """
    for i, t in enumerate(tlist):
        print t
    """
    if ( icheckfile('thomson',100000,1 )):
        print("Exsts")
    else:
        print("Not Exist")
        





