#!/usr/bin/env python
# coding:utf-8
# $Id$

import os
import sys
import re
import random
import shutil
import struct
import threading
import time
from MyViewGlobal import logger, debug

LockCache = threading.Lock()
LockFile = threading.Lock()

class Serializer(object):
    """
    データのシリアライズを行うクラス
    """
    def __init__(self, **args):
        self.datatypechars = ['x', 'c', 'b', 'B', 'h', 'H', 'i', 'I',
                'l', 'L', 'q', 'Q', 'f', 'd', 's', 'p', 'P'] 
        self.datatypedict = {'pad byte':'x',
                'char':'c',
                'signed char':'b',
                'unsigned char':'B',
                'short':'h',
                'unsigned short':'H',
                'int':'i',
                'unsigned int':'I',
                'long':'l',
                'unsigned long':'L',
                'long long':'q',
                'unsigned long long':'Q',
                'float':'f',
                'double':'d',
                'string':'s',
                'char []':'p',
                'void *':'P'}
        self.endian = args.get('endian', '>') # '>' Big endian '<' Little endian

    def pack(self, blocksOfValues):
        """
        オブジェクトの配列をシリアライズしてbytesストリングを返す
        """
        head = b""
        chars = b""
        for dtype, values in blocksOfValues:
            if dtype in self.datatypechars:
                datatype = dtype
            else:
                datatype = self.datatypedict.get(dtype, '')
            l = len(values)
            fmt = "%c%d%c" % (self.endian, l, datatype)
            if ( datatype == 's' ):
                if type(values)==bytes:
                    #print(type(values), values)
                    dat = struct.pack(fmt.encode(), values)
                else:
                    dat = struct.pack(fmt.encode(), values.encode())
            else:
                dat = struct.pack(fmt, *values)
            head = head + b"%c,%d,%d,%c:" % (
                    self.endian.encode(), l, len(dat), datatype.encode())
            chars = chars + dat

        blocksize = len(blocksOfValues)
        headsize = len(head)
        eightbyte = struct.pack(">LL", headsize, blocksize)
        return eightbyte + head + chars
    
    def unpack(self, chars):
        """
        ストリングを配列に戻す
        """
        headsize, blocksize = struct.unpack('>LL', chars[:8])
        info = chars[8: 8+headsize]
        clm = info.split(':')[:-1]
        if blocksize != len(clm):
            return None
        idx = 8 + headsize
        values = []
        for c in clm:
            endian, lstr, size, datatype = c.split(',')
            l = int(lstr)
            size = int(size)
            fmt = "%c%d%c" % (endian, l, datatype)
            string = chars[idx: idx + size]
            idx = idx + size 
            values.append(list(struct.unpack(fmt, string)))
        return values

def RebuildCache(filepath):
    def filename_to_keyname(fname):
        fmt = re.compile(r'(.*)_([0-9]+)_.*')
        m = fmt.match(fname)
        if m:
            diagname, shotno = m.groups()[:2]
        else:
            return None
        return "%s_%s" % (diagname, shotno)


    base, fname = os.path.split(filepath)

    files = os.listdir(base)
    lines = []
    for fn in files:
        #key = SplitName(fn)
        key = filename_to_keyname(fn)
        if key is None:
            continue
        l = "%s\t%s" % (key, fn)
        lines.append(l)
    with open(filepath, 'w') as fp:
        fp.write("\n".join(lines))
        fp.write("\n")                


    base, fname = os.path.split(filepath)

    files = os.listdir(base)
    lines = []
    for fn in files:
        #key = SplitName(fn)
        key = filename_to_keyname(fn)
        if key is None:
            continue
        l = "%s\t%s" % (key, fn)
        lines.append(l)
    with open(filepath, 'w') as fp:
        fp.write("\n".join(lines))
        fp.write("\n")                


class MyViewCache(object):
    """
    Usage In Main App:
        cache = MyViewCache('cache\\cache.txt')
    Usage In Client:
        filepath = cache.find(key)
        if filepath:
            (please use filepath)
        else:
            (There is no file in the cache)
        
        If you want to regist  a file
            cache.registFile(key', TheFilepath)
    """
    def __init__(self, cachefilepath, tempfolder=None):
        # filepath : path to dictionary file in the cache folder
        self.changePath(cachefilepath)
        self.serializer = Serializer()
        if tempfolder:
            self.tempfolder = tempfolder
        else:
            self.tempfolder = os.getcwd()
        
    def registData(self, key, data):
        """
        registData(key, data)
        data = [['double', list_values], ['int', list_integers]]
        """
        line = self.serializer.pack(data)
        filename = '%s.dat' % key
        #filename = self.makeUniqueFilename(filename)
        filename = self.makeUniqueFilename(filename, self.tempfolder)
        #filepath = os.path.join(self.cachefolderpath, filename)
        filepath = os.path.join(self.tempfolder, filename)

        try:
            while not self.lock_file(filepath):
                logger.debug("[registData datafile] Wait unlock the file %s" % filepath)
                time.sleep(0.5)
            fpo = open(filepath, 'wb')
            fpo.write(line)
            fpo.close()
        finally:
            self.unlock_file(filepath)

        tempfilepath = self.registFile(key, filepath) 
        self.releasefile(tempfilepath)
        
        LockCache.acquire() #---------------------------------------------------------
        if key in self.cachedic.keys():
            self.remove(key)

        self.cachedic[key] = filename

        try:
            while not self.lock_file(self.cachefilepath):
                logger.debug("[registData cachefile] Wait unlock the file %s" % self.cachefilepath)
                time.sleep(0.5)
            fpo = open(self.cachefilepath, 'a')
            l = "%s\t%s\n" % (key, filename)
            fpo.write(l)
            fpo.close()
        finally:
            self.unlock_file(self.cachefilepath)
        LockCache.release() #---------------------------------------------------------

        return 

    def restoreData(self, key):
        """
        データの復元
        """
        LockCache.acquire() #---------------------------------------------------------
        filename = self.cachedic.get(key, None)
        LockCache.release() #---------------------------------------------------------

        if filename:
            filepath = os.path.join(self.cachefolderpath, filename)
            try:
                while not self.lock_file(filepath):
                    logger.debug("[restoreData data] Wait unlock the file %s" % filepath)
                    time.sleep(0.5)
                fpi = open(filepath, 'rb')
                line = fpi.read()
                fpi.close()
            finally:
                self.unlock_file(filepath)
            data = self.serializer.unpack(line)
            return data
        else:
            return None
                
    def changePath(self, cachefilepath):
        logger.debug("in changePath")
        #print ">>>>>>>> in changePath"
        self.cachefilepath = cachefilepath
        self.cachefolderpath, self.cachefilename  = os.path.split(self.cachefilepath)
        if self.cachefolderpath == '':
            self.cachefolderpath = '.'

        LockCache.acquire() #--------------------------------------------------------
        self.cachedic = {}

        self.create()
        self.open_cache()
        removes = []
        
        for key in self.cachedic.keys():
            filename = self.cachedic[key]
            filepath = os.path.join(self.cachefolderpath, filename)
            if os.path.exists(filepath):
                pass
            else:
                removes.append(key)

        for key in removes:
            del (self.cachedic[key])
        self.save_cache()
        LockCache.release() #--------------------------------------------------------
        logger.debug("out changePath")
        #print ">>>>>>> out changePath"

    def registFile(self, key, filepath):
        LockCache.acquire() #--------------------------------------------------------
        if key in self.cachedic.keys():
            self.remove(key)
        fpath, fname = os.path.split(filepath)
        fname = self.makeUniqueFilename(fname)
        distpath = os.path.join(self.cachefolderpath, fname)
#        print "copy from %s to %s" % (filepath, distpath)
        logger.debug("move from %s to %s" % (filepath, distpath))
        try:
            while not self.lock_file(filepath):
                logger.debug("[registFile filecopy] Wait unlock the file %s" % filepath)
                time.sleep(0.5)
#            self.copyfile(filepath, distpath)
#            self.removefile(filepath)
            shutil.move(filepath, distpath)
        finally:
            self.unlock_file(filepath)

        self.cachedic[key] = fname
        LockCache.release() #--------------------------------------------------------

        try:
            while not self.lock_file(self.cachefilepath):
                logger.debug("[registFile cachefile] Wait unlock the file %s" % self.cachefilepath)
                time.sleep(0.5)
            fpo = open(self.cachefilepath, 'a')
            l = "%s\t%s\n" % (key, fname)
            fpo.write(l)
            fpo.close()
        finally:
            self.unlock_file(self.cachefilepath)
        
        distpath = self.acquirefile(distpath)
        return distpath

    def removeall(self):
        LockCache.acquire() #-------------------------------------------------------
        for key in self.cachedic.keys():
            filepath = os.path.join(self.cachefolderpath, self.cachedic[key])
            self.removefile(filepath)
        # make cachedic empty
        self.cachedic = {}
        # make empty file
        try:
            while not self.lock_file(self.cachefilepath):
                logger.debug("[removeall] Wait unlock the file %s" % filepath)
                time.sleep(0.5)
            fpo = open(self.cachefilepath, 'w')
            fpo.close()
        finally:
            self.unlock_file(self.cachefilepath)
        LockCache.release() #-------------------------------------------------------

    def find(self, key):
        LockCache.acquire() #------------------------------------------------------
        if key in self.cachedic.keys():
            filepath = os.path.join(self.cachefolderpath, self.cachedic[key])
            if os.path.exists(filepath):
                filepath = self.acquirefile(filepath)
            else:
                filepath = None
        else:
            filepath = None
        LockCache.release() #------------------------------------------------------
        return filepath

    def show_all(self):
        print("---- Data Local File Cache List (%d) ----" % len(self.cachedic), id(self.cachedic))
        print(" CASHE FILE: %s" % self.cachefilepath)
        keys = [k for k in self.cachedic.keys()]
        keys.sort()
        print("\n".join([k for k in keys]))
        print("====================================")


    def findCacheAndTime(self, key):
        LockCache.acquire() #------------------------------------------------------
        if key in self.cachedic.keys():
            filepath = os.path.join(self.cachefolderpath, self.cachedic[key])
            if os.path.exists(filepath):
                filetime = os.stat(filepath).st_mtime
                filepath = self.acquirefile(filepath)
            else:
                filepath = None
                filetime = None
        else:
            filepath = None
            filetime = None
        LockCache.release() #------------------------------------------------------
        return [filepath, filetime]

    def findTime(self, key):
        LockCache.acquire() #------------------------------------------------------
        if key in self.cachedic.keys():
            filepath = os.path.join(self.cachefolderpath, self.cachedic[key])
            if os.path.exists(filepath):
                filetime = os.stat(filepath).st_mtime
            else:
                filetime = None
        else:
            filetime = None
        LockCache.release() #------------------------------------------------------
        return filetime


#----------------------------------------------------------------------------
    def opencache(self):
        LockCache.acquire() #--------------------------------------------------------
        self.open_cache()
        LockCache.release() #--------------------------------------------------------
    
    def savecache(self):
        LockCache.acquire() #--------------------------------------------------------
        self.save_cache()
        LockCache.release() #--------------------------------------------------------

    def open_cache(self):
        logger.debug("in open_cache")
        try:
            while not self.lock_file(self.cachefilepath):
                logger.debug("[open_cache] Wait unlock the file %s" % self.cachefilepath)
                time.sleep(0.5)

            fpi = open(self.cachefilepath, 'r')
            lines = fpi.readlines()
            fpi.close()
        except:
            logger.debug("Error in open cache file %s." % self.cachefilepath)
        finally:
            self.unlock_file(self.cachefilepath)
        logger.debug(">> %s %d %s" % (self.cachefilepath, len(lines), id(self.cachedic)))
       
        current_cached_data_in_file = {}
        for l in lines:
            l = l.strip('\n')
            l = l.strip('\r')
            if len(l) == 0: continue
            if l[0] == '#': continue
            clm = l.split('\t')
            if (len(clm) >= 2 ):
                current_cached_data_in_file[clm[0]] = clm[1]
        self.cachedic.update(current_cached_data_in_file)

    def save_cache(self):
        try:
            while not self.lock_file(self.cachefilepath):
                logger.debug("[save_cache] Wait unlock the file %s" % self.cachefilepath)
                time.sleep(0.5)
            fpo = open(self.cachefilepath, 'w')
            for key in self.cachedic.keys():
                l = "%s\t%s\n" % (key, self.cachedic[key])
                fpo.write(l)
            fpo.close()
        finally:
            self.unlock_file(self.cachefilepath)

    def remove(self, key):
        #filepath = self.cachedic.get(key, None)
        filename = self.cachedic.get(key, None)
        if filename:
            filepath = os.path.join(self.cachefolderpath, filename)
            self.removefile(filepath)
            del self.cachedic[key]
            self.save_cache()


    def create(self):
        # check cache folder or create
        logger.debug("in create")
        if os.path.exists(self.cachefolderpath):
            pass
        else:
            os.makedirs(self.cachefolderpath)

        # check cache dictionary file or create
        if os.path.exists(self.cachefilepath):
            pass
        else:
            if not self.lock_file(self.cachefilepath):
                # Lock file exists = cache file is exists."
                pass
            else:
                logger.debug("cachefile is locked !")
                fpo = open(self.cachefilepath, 'w')
                fpo.close()
                self.unlock_file(self.cachefilepath)
                logger.debug("cachefile is unlocked !")
        logger.debug("out creat")

    def makeUniqueFilename(self, filename, folderpath=None):
        """
        ユニークなファイル名をつける
        """
        if not folderpath:
            folderpath = self.cachefolderpath
        base, ext = os.path.splitext(filename)
        filenames = os.listdir(folderpath)
        filename = base + '_%s' % self.randomstr() + ext
        #print filename
        while (filename in filenames):
            #print "Retry"
            filename = base + '_%s' % self.randomstr() + ext
            logger.debug(filename)
        return filename

    def randomstr(self, length=10):
        """
        ランダムな文字列を返す
        """
        ret = ""
        source = "abcdefghijklmnopqrstuvwxyz0123456789"
        source = source + source.upper()
        for i in range(length):
            ret += random.choice(source)
        return ret

    def acquirefile(self, filepath):
        logger.debug(">>>> ACQUIRE %s" % filepath)
        path, filename = os.path.split(filepath)
        filename = self.makeUniqueFilename(filename, self.tempfolder)
        distpath = os.path.join(self.tempfolder, filename)
        timeout = 10
        try:
            while not self.lock_file(filepath) and timeout > 0:
                logger.debug("[acquirefile] Wait unlock the file %s" % filepath)
                time.sleep(0.5)
                timeout -= 1
            self.copyfile(filepath, distpath)
        finally:
            self.unlock_file(filepath)
        return distpath

    def releasefile(self, filepath):
        """
        ファイルの削除
        """
        self.removefile(filepath)

    def copyfile(self, src, dst):
        if not os.path.exists(src):
            return
        fsize = os.stat(src).st_size
        fsrc = open(src, "rb")
        fdst = open(dst, "wb")
        while True: 
            buf = fsrc.read(2**20)
            if not buf:
                break
            fdst.write(buf)
        if fdst:
            fdst.close()
        if fsrc:
            fsrc.close()

    def removefile(self, filepath):
        if os.path.exists(filepath):
            timeout = 10
            keep = True
            while keep and timeout > 0:
                try:
                    os.unlink(filepath)
                    keep = False
                except:
                    logger.debug("Wait removed %s ..." % filepath)
                    time.sleep(0.5)
                if not os.path.exists(filepath):
                    keep = False
                timeout -= 1

    def lock_file(self, filepath):
        """
        ロックファイルを作成
        """
        path, filename = os.path.split(filepath)
        base, ext = os.path.splitext(filename)
        lockfilename = base+'.loc'
        lockfilepath = os.path.join(path, lockfilename)
        with LockFile:
            if os.path.exists(lockfilepath):
                return False
            else:
                try:
                    fpo = open(lockfilepath, "w")
                    fpo.close()
                except:
                    return False
                return lockfilepath
            
    def unlock_file(self, filepath):
        """
        ロックファイルを削除
        """
        path, filename = os.path.split(filepath)
        base, ext = os.path.splitext(filename)
        lockfilename = base+'.loc'
        lockfilepath = os.path.join(path, lockfilename)
        with LockFile:
            if os.path.exists(lockfilepath):
                try:
                    self.removefile(lockfilepath)
                except:
                    import traceback
                    if debug:
                        logger.error('Fail to remove '+filepath)
                        logger.error(traceback.print_exc())
                    return False

if __name__ == '__main__':
    ca = MyViewCache('cache/cache.txt')
    ca.registFile('test', 'test.ps')
#    ca.registFile('test1', 'test.ps')
#    ca.registFile('test2', 'test.ps')
#    ca.registFile('test3', 'test.ps')
#    ca.registFile('test4', 'test.ps')
#    ca.registFile('test5', 'test.ps')
    print(ca.cachedic)
    #ca.removeall()
    #print ca.cachedic

